//////////////////////////////////
//		TI-BASIC COMPILER		//
//		  Matthew Iselin		//
//////////////////////////////////
// Because I'm sick of using	//
// the calculator's keypad		//
// to write programs...			//
//////////////////////////////////
// This is the core of the		//
// compiler.					//
//////////////////////////////////
// Created: 15-5-07				//
// License: GPL					//
// Last Modified: 25-5-07		//
//////////////////////////////////

// TODOS /////////////////////////
// TODO: finish symbol tables (Stat vars only)
// TODO: finish definitions in tokens.h (See above)

// DEFINES ///////////////////////
#ifndef _WIN32
#define HIWORD(var) (((var) & 0xffff0000) >> 16)
#define LOWORD(var) ((var) & 0x0000ffff)

#define LOBYTE(var) ((var) & 0x00ff)
#define HIBYTE(var) (((var) & 0xff00) >> 8)
#endif

#ifdef _DEBUG
#define DEBUG					// this is for MSVC++ 6 compatibility (it sets _DEBUG for debug versions)
								// debugging just spews all sorts of
								// information while the compiler runs
#endif

// INCLUDES //////////////////////
#ifdef _WIN32
#include <windows.h>
#endif // #ifdef _WIN32
#include <iostream>
#include <string.h>
#include <fstream>
#include <stdio.h>
#include <string>
#include <vector>
#include <time.h>
#include "Duration.h"
#include "tokens.h"

// USING DECLARATIONS ////////////
using namespace std;

// TYPES /////////////////////////

// a token
struct Token {
	unsigned char data;	// to be written to the file
	char* text;			// what the TI-BASIC equivalent is
};

// a two byte variable (0xBB, 0x7E and SysVar)
struct TwoByte {
	unsigned short data; // token
	char* text; // ti-basic equivalent
};

// a rule to replace an ascii character to a token
struct ConvertRule {
	char c;				// the character
	unsigned char tok;	// the equivalent token
};

// program header
struct PrgmHeader {
	char ProgSig[8]; // ALWAYS **TI83F*
	char ExtSig[3]; // always 0x1A, 0x0A, 0x00
	char Comment[42]; // file comment
	unsigned short DataSize; // size of the data section
};

// a variable entry
struct VarEnt {
	unsigned short VarStart; // always 0xB or 0xD (means we can find the start of the entry)
	unsigned short VarLength1; // length of the data
	unsigned char VarType; // variable type, always 0x05
	char VarName[8]; // padded with spaces
	char Version; // usually 0
	char Flag; // 0x80 if archived
	unsigned short VarLength2; // same as above
};

// HANDLER PROTOTYPES ////////////
void TokenHandler( unsigned char data, vector<string> str );

// GLOBAL VARIABLES //////////////

// timer
CDuration myDuration;

// verbose or not?
bool VerboseMode = false;

// are we decompiling or compiling?
bool Compiling = true;

// assembly?
bool assembly = false;

// program code
string progData = "";

// Standard Tokens
// Standard Tokens are any token that can be used anywhere in the line.
struct Token StandardTokens[] = {
	/** CONTROL page of PROGRAM EDITOR (press PRGM when EDITING a program) **/
	{ LOGIC_IF,			"If "		},
	{ LOGIC_THEN,		"Then"		},
	{ LOGIC_ELSE,		"Else"		},
	{ CTL_FOR,			"For("		},
	{ CTL_WHILE,		"While "	},
	{ CTL_REPEAT,		"Repeat "	},
	{ CTL_END,			"End"		},
	{ CTL_PAUSE,		"Pause "	},
	{ LABEL,			"Label "	},
	{ CTL_GOTO,			"Goto "		},
	{ INCSKIPIFHIGH,	"IS>("		},
	{ DECSKIPIFLOW,		"DS<("		},
	{ MENU,				"Menu("		},
	{ PROGRAM,			"prgm"		},
	{ CTL_RETURN,		"Return"	},
	{ CTL_STOP,			"Stop"		},

	/** I/O page **/
	{ INPUT,			"Input "	},
	{ PROMPT,			"Prompt "	},
	{ DISP,				"Disp "		},
	{ DISPGRAPH,		"DispGraph"	},
	{ DISPTABLE,		"DispTable"	},
	{ OUTPUT,			"Output("	},
	{ GETKEY,			"getKey"	},
	{ CLRHOME,			"ClrHome"	},
	{ CLRTABLE,			"ClrTable"	},

	/** GetCalc **/
	{ GET,				"Get("		},
	{ SEND,				"Send("		},

	/** Others **/
	{ TRACE,			"Trace"		},
	{ ZSTANDARD,		"ZStandard"	},
	{ ZTRIG,			"ZTrig"		},
	{ ZBOX,				"ZBox"		},
	{ ZOOMIN,			"ZoomIn"	},
	{ ZOOMOUT,			"ZoomOut"	},
	{ ZSQUARE,			"ZSquare"	},
	{ ZINTEGER,			"ZInteger"	},
	{ ZPREVIOUS,		"ZPrevious"	},
	{ ZDECIMAL,			"ZDecimal"	},
	{ ZOOMSTAT,			"ZoomStat"	},
	{ ZOOMRCL,			"ZoomRcl"	},
	{ ZOOMSTO,			"ZoomSto"	},
	{ TEXTFUNC,			"Text("		},
	{ STOREPIC,			"StorePic"	},
	{ RECALLPIC,		"RecallPic"	},
	{ STOREGDB,			"StoreGDB"	},

	/** OTHERS **/
	{ TO_DMS,			"->DMS"			},
	{ TO_DEC,			"->DEC"			},
	{ TO_FRAC,			"->FRAC"		},
	{ STORE,			"->"			},
	{ BOXPLOT,			"BoxPlot"		},
	{ RADIANS,			"[radians]"		},
	{ DEGREES,			"[degrees]"		},
	{ INVERSE,			"^-1"			},
	{ SQUARE,			"^2"			},
	{ TRANSPOSE,		"[transpose]"	},
	{ CUBE,				"^3"			},
	{ ROUND,			"round("		},
	{ PXLTEST,			"pxl-Test("		},
	{ AUGMENT,			"augment("		},
	{ ROWSWAP,			"RowSwap("		},
	{ ROWPLUS,			"row+("			},
	{ STARROW,			"*row("			},
	{ STARROWPLUS,		"*row+("		},
	{ MAX,				"max("			},
	{ MIN,				"min("			},
	{ MEDIAN,			"median("		},
	{ RANDM,			"randM("		},
	{ MEAN,				"mean("			},
	{ SOLVE,			"solve("		},
	{ SEQFUNC,			"seq("			},
	{ FNINT,			"fnInt("		},
	{ NDERIV,			"NDeriv("		},
	{ FMIN,				"FMin("			},
	{ FMAX,				"FMax("			},
	{ CUBICREG,			"CubicReg "		},
	{ QUARTREG,			"QuartReg "		},
	{ LOGIC_OR,			" or "			},
	{ LOGIC_XOR,		" xor "			},
	{ LOGIC_AND,		"and"			},
	{ STR_THETA,		"[theta]"		},
	{ PROGRAM,			"prgm"			},
	{ RADIAN,			"Radian"		},
	{ DEGREE,			"Degree"		},
	{ NORMAL,			"Normal"		},
	{ SCI,				"Sci"			},
	{ ENG,				"Eng"			},
	{ FLOAT,			"Float"			},
	{ TEST_LOREQU,		"<="			},
	{ TEST_GOREQU,		">="			},
	{ TEST_NOTEQUAL,	"!="			},
	{ ANSWER,			"Ans"			},
	{ FIX,				"Fix "			},
	{ HORIZ,			"Horiz"			},
	{ FULL,				"Full"			},
	{ FUNC,				"Func"			},
	{ PARAM,			"Param"			},
	{ POLAR,			"Polar"			},
	{ SEQ,				"Seq"			},
	{ INDPNTAUTO,		"IndpntAuto"	},
	{ INDPNTASK,		"IndpntAsk"		},
	{ DEPENDAUTO,		"DependAuto"	},
	{ DEPENDASK,		"DependAsk"		},
	{ BOX,				"[box]"			},
	{ DOT,				"[dot]"			},
	{ TRACE,			"Trace"			},
	{ CLRDRAW,			"ClrDraw"		},
	{ ZSTANDARD,		"ZStandard"		},
	{ ZTRIG,			"ZTrig"			},
	{ ZBOX,				"ZBox"			},
	{ ZOOMIN,			"Zoom In"		},
	{ ZOOMOUT,			"Zoom Out"		},
	{ ZSQUARE,			"ZSquare"		},
	{ ZINTEGER,			"ZInteger"		},
	{ ZPREVIOUS,		"ZPrevious"		},
	{ ZDECIMAL,			"ZDecimal"		},
	{ ZOOMSTAT,			"ZoomStat"		},
	{ ZOOMRCL,			"ZoomRcl"		},
	{ PRINTSCREEN,		"PrintScreen"	},
	{ ZOOMSTO,			"ZoomSto"		},
	{ TEXTFUNC,			"Text("			},
	{ NPR,				"nPr"			},
	{ NCR,				"nCr"			},
	{ FNON,				"FnOn "			},
	{ FNOFF,			"FnOff "		},
	{ STOREPIC,			"StorePic "		},
	{ RECALLPIC,		"RecallPic "	},
	{ STOREGDB,			"StoreGDB "		},
	{ RECALLGDB,		"RecallGDB "	},
	{ LINE,				"Line("			},
	{ VERTICAL,			"Vertical "		},
	{ PTON,				"Pt-On("		},
	{ PTOFF,			"Pt-Off("		},
	{ PTCHANGE,			"Pt-Change("	},
	{ PXLON,			"Pxl-On("		},
	{ PXLOFF,			"Pxl-Off("		},
	{ PXLCHANGE,		"Pxl-Change("	},
	{ SHADE,			"Shade("		},
	{ CIRCLE,			"Circle("		},
	{ HORIZONTAL,		"Horizontal "	},
	{ TANGENT,			"Tangent("		},
	{ DRAWINV,			"DrawInv "		},
	{ DRAWF,			"DrawF "		},
	{ RAND,				"rand"			},
	{ PI,				"[pi]"			},
	{ GETKEY,			"getKey"		},
	{ NEGATIVE,			"[neg]"			},
	{ CONV_INT,			"int("			},
	{ ABS,				"abs("			},
	{ DETERMINANT,		"det("			},
	{ IDENTITY,			"identity("		},
	{ DIM,				"dim("			},
	{ SUM,				"sum("			},
	{ PROD,				"prod("			},
	{ NOT,				"not("			},
	{ IPART,			"iPart("		},
	{ FPART,			"fPart("		},
	{ SQR_ROOT,			"[root]^2"		},
	{ CUBE_ROOT,		"[root]^3"		},
	{ NATLOG,			"ln("			},
	{ ETOPOWER,			"e^"			},
	{ LOGARITHM,		"log("			},
	{ POWER10,			"10^("			},
	{ SINE,				"sin("			},
	{ INVSIN,			"asin("			},
	{ COSINE,			"cos("			},
	{ INVCOSINE,		"acos("			},
	{ TANGENT_TRIG,		"tan("			},
	{ INVTANGENT,		"atan("			},
	{ HYP_SINE,			"sinh("			},
	{ HYP_ISINE,		"asinh("		},
	{ HYP_COSINE,		"cosh("			},
	{ HYP_ICOSINE,		"acosh("		},
	{ HYP_TANGENT,		"tanh("			},
	{ HYP_ITANGENT,		"atanh("		},
	{ LOGIC_ELSE,		"Else"			},
	{ FILL,				"Fill("			},
	{ SORTA,			"SortA("		},
	{ SORTD,			"SortD("		},
	{ LIST,				"[list]"		},
	{ PLOT1,			"Plot1("		},
	{ PLOT2,			"Plot2("		},
	{ PLOT3,			"Plot3("		},
	{ XTHROOT,			"[root]^"		},
	{ VARSTATS_1,		"1-Var Stats "	},
	{ VARSTATS_2,		"2-Var Stats "	},
	{ LINREG1,			"LinReg(a+bx) "	},
	{ EXPREG,			"ExpReg "		},
	{ LNREG,			"LnReg "		},
	{ PWRREG,			"PwrReg "		},
	{ MEDMED,			"Med-Med "		},
	{ QUADREG,			"QuadReg "		},
	{ CLRLIST,			"ClrList "		},
	{ CLRTABLE,			"ClrTable"		},
	{ HISTOGRAM,		"Histogram"		},
	{ XYLINE,			"xyLine"		},
	{ SCATTER,			"Scatter"		},
	{ LINREG2,			"LinReg(ax+b) "	},
};

// two-byte variables
struct TwoByte CalcVars[] = {
	// AsmPrgm (uncompiled)
	{ 0x6CBB,			"AsmPrgm"	},
	{ 0x6DBB,			"AsmPrgm"	}, // this means decompilation works, but compilation won't hit this

	// SysVar
	{ MAT_A,			"[A]"		},
	{ MAT_B,			"[B]"		},
	{ MAT_C,			"[C]"		},
	{ MAT_D,			"[D]"		},
	{ MAT_E,			"[E]"		},
	{ MAT_F,			"[F]"		},
	{ MAT_G,			"[G]"		},
	{ MAT_H,			"[H]"		},
	{ MAT_I,			"[I]"		},
	{ MAT_J,			"[J]"		},
	{ L1,				"L1"		},
	{ L2,				"L2"		},
	{ L3,				"L3"		},
	{ L4,				"L4"		},
	{ L5,				"L5"		},
	{ L6,				"L6"		},
	{ L7,				"L7"		},
	{ L8,				"L8"		},
	{ L9,				"L9"		},
	{ L0,				"L0"		},
	{ Y1,				"Y1"		},
	{ Y2,				"Y2"		},
	{ Y3,				"Y3"		},
	{ Y4,				"Y4"		},
	{ Y5,				"Y5"		},
	{ Y6,				"Y6"		},
	{ Y7,				"Y7"		},
	{ Y8,				"Y8"		},
	{ Y9,				"Y9"		},
	{ Y0,				"Y0"		},
	{ X1T,				"X1T"		},
	{ Y1T,				"Y1T"		},
	{ X2T,				"X2T"		},
	{ Y2T,				"Y2T"		},
	{ X3T,				"X3T"		},
	{ Y3T,				"Y3T"		},
	{ X4T,				"X4T"		},
	{ Y4T,				"Y4T"		},
	{ X5T,				"X5T"		},
	{ Y5T,				"Y5T"		},
	{ X6T,				"X6T"		},
	{ Y6T,				"Y6T"		},
	{ R1,				"R1"		},
	{ R2,				"R2"		},
	{ R3,				"R3"		},
	{ R4,				"R4"		},
	{ R5,				"R5"		},
	{ R6,				"R6"		},
	{ PIC1,				"PIC1"		},
	{ PIC2,				"PIC2"		},
	{ PIC3,				"PIC3"		},
	{ PIC4,				"PIC4"		},
	{ PIC5,				"PIC5"		},
	{ PIC6,				"PIC6"		},
	{ PIC7,				"PIC7"		},
	{ PIC8,				"PIC8"		},
	{ PIC9,				"PIC9"		},
	{ PIC0,				"PIC0"		},
	{ GDB1,				"GDB1"		},
	{ GDB2,				"GDB2"		},
	{ GDB3,				"GDB3"		},
	{ GDB4,				"GDB4"		},
	{ GDB5,				"GDB5"		},
	{ GDB6,				"GDB6"		},
	{ GDB7,				"GDB7"		},
	{ GDB8,				"GDB8"		},
	{ GDB9,				"GDB9"		},
	{ GDB0,				"GDB0"		},
	{ STR1,				"STR1"		},
	{ STR2,				"STR2"		},
	{ STR3,				"STR3"		},
	{ STR4,				"STR4"		},
	{ STR5,				"STR5"		},
	{ STR6,				"STR6"		},
	{ STR7,				"STR7"		},
	{ STR8,				"STR8"		},
	{ STR9,				"STR9"		},
	{ STR0,				"STR0"		},
	// 7E Variables
	{ SEQ,				"Sequential"	},
	{ SIMUL,			"Simul"			},
	{ POLARGC,			"PolarGC"		},
	{ RECTGC,			"RectGC"		},
	{ COORDON,			"CoordOn"		},
	{ COORDOFF,			"CoordOff"		},
	{ CONNECTED,		"Connected"		},
	{ DOT_7E,			"Dot"			},
	{ AXESON,			"AxesOn"		},
	{ AXESOFF,			"AxesOff"		},
	{ GRIDON,			"GridOn"		},
	{ GRIDOFF,			"GridOff"		},
	{ LABELON,			"LabelOn"		},
	{ LABELOFF,			"LabelOff"		},
	{ WEB,				"Web"			},
	{ TIME,				"Time"			},
	{ UVAXES,			"uvAxes"		},
	{ VWAXES,			"vwAxes"		},
	{ UWAXES,			"uwAxes"		},
	// BB Variables
	{ NPV,				"npv("			},
	{ IRR,				"irr("			},
	{ BAL,				"bal("			},
	{ SUMPRN,			"sum_prn("		},
	{ SUMINT,			"sum_int("		},
	{ TONOM,			"->Nom"			},
	{ TOEFF,			"->Eff"			},
	{ DBD,				"dbd("			},
	{ LCM,				"lcm("			},
	{ GCD,				"gcd("			},
	{ RANDINT,			"RandInt("		},
	{ RANDBIN,			"RandBin("		},
	{ SUB,				"sub("			},
	{ STDDEV,			"StdDev("		},
	{ VARIANCE,			"variance("		},
	{ INSTRING,			"inString("		},
	{ NORMALCDF,		"normalcdf("	},
	{ INVNORM,			"invNorm("		},
	{ TCDF,				"tcdf("			},
	{ XSQUCDF,			"x^2cdf("		},
	{ FCDF,				"fcdf("			},
	{ BINOMPDF,			"binompdf("		},
	{ BINOMCDF,			"binomcdf("		},
	{ POISSONPDF,		"poissonpdf("	},
	{ POISSONCDF,		"poissoncdf("	},
	{ GEOMETPDF,		"geometpdf("	},
	{ GEOMETCDF,		"geometcdf("	},
	{ NORMALPDF,		"normalpdf("	},
	{ TPDF,				"tpdf("			},
	{ XSQUPDF,			"x^2pdf("		},
	{ FPDF,				"fpdf("			},
	{ RANDNORM,			"RandNorm("		},
	{ TVM_PMT,			"tvm_pmt"		},
	{ TVM_I_PERCENT,	"tvm_i%"		},
	{ TVM_PV,			"tvm_PV"		},
	{ TVM_N,			"tvm_N"			},
	{ TVM_FV,			"tvm_FV"		},
	{ CONJ,				"conj("			},
	{ REAL,				"real("			},
	{ IMAG,				"imag("			},
	{ ANGLE,			"angle("		},
	{ CUMSUM,			"cumSum("		},
	{ EXPR,				"expr("			},
	{ LENGTH,			"length("		},
	{ DELTA_LIST,		"delta_List("	},
	{ REF,				"ref("			},
	{ RREF,				"rref("			},
	{ TORECT,			"->Rect"		},
	{ TOPOLAR,			"->Polar"		},
	{ VAR_E,			"[e]"			}, // e by itself is impossible, and dangerous (imagine Disp "Hello"!)
	{ SINREG,			"SinReg "		},
	{ LOGISTIC,			"Logistic "		},
	{ LINREGTTEST,		"LinRegTTest "	},
	{ SHADENORM,		"ShadeNorm("	},
	{ SHADE_T,			"Shade_t("		},
	{ SHADEXSQU,		"Shade_x^2("	},
	{ SHADEF,			"ShadeF("		},
	{ MATRTOLIST,		"Matr->list"	},
	{ LISTTOMATR,		"List->matr"	},
	{ ZTEST,			"Z-Test("		},
	{ TTEST,			"T-Test "		},
	{ TWO_SAMPZTEST,	"2-SampZTest("	},
	{ ONE_PROPZTEST,	"1-PropZTest("	},
	{ TWO_PROPZTEST,	"2-PropZTest("	},
	{ XSQUTEST,			"x^2_test("		},
	{ ZINTERVAL,		"ZInterval"		},
	{ TWO_SAMPZINT,		"2-SampZInt("	},
	{ ONE_PROPZINT,		"1-PropZInt("	},
	{ TWO_PROPZINT,		"2-PropZInt("	},
	{ GRAPHSTYLE,		"GraphStyle("	},
	{ TWO_SAMPTTEST,	"2-SampTTest "	},
	{ TWO_SAMPFTEST,	"2-SampFTest_"	},
	{ TINTERVAL,		"TInterval "	},
	{ TWO_SAMPTINT,		"2-SampTInt "	},
	{ SETUPEDITOR,		"SetUpEditor "	},
	{ PMT_END,			"PMT_End"		},
	{ PMT_BGN,			"PMT_Bgn"		},
	{ REAL_BB,			"Real"			},
	{ REPOWTHETAI,		"re^[theta]i"	},
	{ APLUSBI,			"a+bi"			},
	{ EXPRON,			"ExprOn"		},
	{ EXPROFF,			"ExprOff"		},
	{ CLRALLLISTS,		"ClrAllLists"	},
	{ GETCALC,			"GetCalc("		},
	{ DELVAR,			"DelVar "		},
	{ EQUTOSTRING,		"Equ->String("	},
	{ STRINGTOEQU,		"String->Equ("	},
	{ CLEARENTRIES,		"Clear Entries"	},
	{ SELECT,			"Select("		},
	{ ANOVA,			"ANOVA("		},
	{ MODBOXPLOT,		"ModBoxPlot"	},
	{ NORMPROBPLOT,		"NormProbPlot"	},
};

// Replacements
// Replacements are rules that define special characters that must be replaced with a token.
struct ConvertRule Replacements[] = {
	{ '"', DOUBLEQUOTE },
	{ '\'', APOSTROPHE },
	{ ',', COMMA },
	{ '?', QUESTIONMARK },
	{ ' ', SPACE },
	{ '=', TEST_EQUAL },
	{ '<', TEST_LESSTHAN },
	{ '>', TEST_HIGHTHAN },
	{ '+', ADD },
	{ '-', SUBTRACT },
	{ '/', DIVIDE_SLASH },
	{ '*', MULTIPLY },
	{ '!', FACTORIAL },
	{ ':', COLON },
	{ '\n', HARD_RETURN },
	{ '0', NUM_0 },
	{ '1', NUM_1 },
	{ '2', NUM_2 },
	{ '3', NUM_3 },
	{ '4', NUM_4 },
	{ '5', NUM_5 },
	{ '6', NUM_6 },
	{ '7', NUM_7 },
	{ '8', NUM_8 },
	{ '9', NUM_9 },
	{ '.', PERIOD },
	{ '[', LEFTSBRACK },
	{ ']', RIGHTSBRACK },
	{ '{', LEFTBRACE },
	{ '}', RIGHTBRACE },
	{ '(', LEFTBRACKET },
	{ ')', RIGHTBRACKET },
	{ '&', LOGIC_AND },
	{ '|', LOGIC_OR },
	{ '~', LOGIC_XOR },
	{ '=', TEST_EQUAL },
	{ '<', TEST_LESSTHAN },
	{ '>', TEST_HIGHTHAN },
	{ '^', POWEROF },
};

// number of tokens and replacements
int numVars = sizeof( CalcVars ) / sizeof( TwoByte );
int numTokens = sizeof( StandardTokens ) / sizeof( struct Token );
int numReps = sizeof( Replacements ) / sizeof( struct ConvertRule );

// UTILITY FUNCTIONS /////////////

#ifndef _WIN32

int GetFileTitle(const char *source, char *dest, unsigned int destLength) {
	strlcpy(dest, source, destLength);
	return 0;
}
/*int GetFileTitle(const char *source, char *dest, unsigned int destLength) {
	int sourceLength = strlen(source);
	int i;
	
	strlcpy(dest, source, destLength);
	
	for (i = sourceLength - 1; i >= 0; --i) {
		if ('.' == dest[i]) {
			dest[i] = 0;
			break;
		}
	}
	
	return 0;
}*/

#endif


// progChecksum: calculates the checksum of the data given
// this must be done on an unsigned type, because of the
// limits of the signed type
string progChecksum( const unsigned char* myData, int size )
{
	// return value
	int ret = 0;

	// calculate
	for( int i = size; i >= 0; i-- )
	{
		ret += myData[i];
	}

	// modulus
	ret %= 65536;

	// put into a short value
	short chk = HIWORD( ret ) + LOWORD( ret );

	// return string value
	string s = "";
	s += LOBYTE( chk ); // low 8 bytes
	s += HIBYTE( chk ); // hi 8 bytes

	// return it
	return s;
}

// converts a string to upper case
char* strtoupper( char* str )
{
	for( int i = 0; i < strlen( str ); i++ )
	{
		if( ! ( isupper( str[i] ) ) && isalpha( str[i] ) )
			str[i] = _toupper( str[i] );
	}
	return str;
}

// outputs information about the compilation process
void DisplayInfo( char* str )
{
	if( VerboseMode )
		cout << "[STATUS::" << myDuration.GetDurationLive() << "] " << str << endl;
}

// displays an error message
#define DisplayError(str) cerr << "[ERROR in " << __func__ << ":" << __LINE__ << " ::" << myDuration.GetDurationLive() << "] " << str << endl

// handles the input line
void HandleInput( string in )
{
	// offset
	int off = 0;

	// index
	int i = 0;

	// now work on each character since
	for( int z = off; z < in.length() && off < in.length(); z++ )
	{
		// work on each variable
		for( i = 0; i < numVars; i++ )
		{
			// check for a match
			if( in.substr( off, strlen( CalcVars[i].text ) ) == CalcVars[i].text )
			{
				// match, handle it
				progData += LOBYTE( CalcVars[i].data );
				progData += HIBYTE( CalcVars[i].data );

#ifdef DEBUG
				// print out some information
				printf( "Found match: %s\n", CalcVars[i].text );
#endif

				// increment the offset
				off += strlen( CalcVars[i].text );

				// one has been found...
				break;
			}
		}

		// see if something was replaced, if so, keep going
		if( i != numVars )
			continue;

		// now work on each standard token
		for( i = 0; i < numTokens; i++ )
		{
			// check for a match
			if( in.substr( off, strlen( StandardTokens[i].text ) ) == StandardTokens[i].text )
			{
				// match, handle it
				progData += StandardTokens[i].data;

#ifdef DEBUG
				// print out some information
				printf( "Found match: %s\n", StandardTokens[i].text );
#endif

				// increment the offset
				off += strlen( StandardTokens[i].text );

				// one has been found...
				break;
			}
		}

		// check if none was replaced
		if( i == numTokens )
		{
			int y;
			
			// handle replacements for this character
			for( y = 0; y < numReps; y++ )
			{
				// does it match
				if( (unsigned char) in[off] == Replacements[y].c )
				{
					// we have a match, so handle it
					progData += Replacements[y].tok;

#ifdef DEBUG
					// print out some information
					printf( "Replacing: %c\n", Replacements[y].c );

					// this is really only useful if you are debugging
					in[off] = (unsigned char) Replacements[y].tok;
#endif

					// break from the loop
					break;
				}
			}

			// check if we need to append plaintext
			if( y == numReps )
			{
				// append the data
				progData += isupper( in[off] ) ? in[off] : toupper( in[off] );
			}

			// always increment the offset after a replacement
			off++;
		}
	}
}

// searches the argv array for a given string
int SearchArgv( char* argv[], int argc, char* str )
{
	// search
	for( int i = 0; i < argc; i++ )
	{
		if( strcmp( argv[i], str ) == 0 )
			return i;
	}

	// return it
	return false;
}

// removes [ext] from the end of the string
string RemoveExtension( string fn, string ext )
{
	// see what the last 4 characters are
	if( fn.substr( fn.length() - 4 ) == ext )
	{
		// good to delete
		fn.replace( fn.length() - 4, 4, "" );

		// return the (modified) string
		return fn;
	}
	else
	{
		// not a file of that type
		return ".";
	}
}

// removes the filename from the end of the string
string RemoveFilename( string fn )
{
	int i;
	
	// go from the back to find the index of the last slash
	for( i = fn.length(); i > 0; i-- )
	{
		if( fn[i] == '/' || fn[i] == '\\' )
			break;
	}

	// check if one was found
	if( i == 0 )
		return "";

	// now substring it
	return fn.substr( 0, i + 1 ); // KEEP the slash
}

// FUNCTIONS /////////////////////

// main: entry point to application
// at the moment no 'link', so only one
// file at a time...
int main( int argc, char* argv[] )
{
	// error string
	string errstr;

	// splash banner
	cout << "--------------- TI-BASIC COMPILER ---------------" << endl;
	cout << "          Copyright 2007 Matthew Iselin          " << endl;
	cout << "-------------------------------------------------" << endl;

	// check for valid number of arguments
	if( argc < 2 || argv[1] == NULL )
	{
		// tell the user
		errstr = "Usage: tibasic.exe filename [options]\nOptions:\n\t-v:\tVerbose mode";
		DisplayError( const_cast< char* >( errstr.c_str() ) );
		return -1;
	}

	// look for special arguments
	if( argc > 2 )
	{
		// temporary storage
		int tmp = 0;

		// check for verbose
		if( tmp = SearchArgv( argv, argc, "-v" ) )
		{
			// enable verbose mode
			VerboseMode = true;
		}

		// check for weird way of doing switches
		if( tmp = SearchArgv( argv, argc, "/v" ) )
		{
			// enable verbose mode
			VerboseMode = true;
		}

		// check for decompilation
		if( tmp = SearchArgv( argv, argc, "-d" ) )
		{
			// decompile
			Compiling = false;
		}

		// check for weird way of doing switches
		if( tmp = SearchArgv( argv, argc, "/d" ) )
		{
			// decompile
			Compiling = false;
		}
	}

	// allocate space for the file title
	char* fileTitle = (char*) malloc( strlen( argv[1] ) + 1 );

	// get the file title (ie. no extension)
	int err = GetFileTitle( argv[1], fileTitle, strlen( argv[1] ) + 1 );
	
	printf("fileTitle = %s\n", fileTitle);
	
	// check for valid title
	if( err == -1 )
	{
		// tell the user
		errstr = string( "Invalid filename: " ) + string( argv[1] );
		DisplayError( const_cast< char* >( errstr.c_str() ) );

		// free the memory we took
		free( fileTitle );

		// return an error
		return -1;
	}

	// are we compiling or decompiling?
	if( Compiling )
	{
		// the destination filename
		string destFile = RemoveExtension( fileTitle, ".tib" ) + ".8xp";

		// the destination path
		string destPath = RemoveFilename( argv[1] );

		// check if it's valid
		if( destFile == "..8xp" )
		{
			// tell the user
			errstr = "Invalid filename: " + string( argv[1] );
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}

		// and now truncate fileTitle so we have a program name
		string ProgName = strtoupper( const_cast< char* >( RemoveExtension( fileTitle, ".tib" ).substr( 0, 8 ).c_str() ) );

		// pad with '/' characters (this way they can be converted to NULLs)
		for( int padCount = ProgName.length(); padCount < 8; padCount++ )
			ProgName += '/';

		// good filename
		cout << "Compiling " << argv[1] << "..." << endl;

		// open the file
		ifstream ifs;
		ifs.open( argv[1] );

		// check that it's valid
		if( ifs.bad() || ifs.fail() )
		{
			// tell the user
			errstr = "Couldn't open " + string( argv[1] );
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}

		// tell the user what's happening
		DisplayInfo( "Parsing file..." );

		// parse each line
		while( ! ( ifs.eof() ) )
		{
			// temporary storage
			string tmp;

			// read in a set of instructions
			getline( ifs, tmp, '\n' );

			// check if it's valid (this might happen with a blank line)
			if( tmp == "" ) { continue; };

			// handle this line
			HandleInput( tmp );

			// append the newline
			progData += (unsigned char) HARD_RETURN;
		}

		// tell the user what's happening
		DisplayInfo( "Building output file..." );

		// length
		unsigned short len;

		// parsing is done, now setup variable header
		string varHeader = "";
		len = progData.length();
		varHeader += LOBYTE( len );
		varHeader += HIBYTE( len );

		// now setup the data section header (variable entry)
		string varEnt = "";
		varEnt += 0x0D;
		varEnt += '\0';
		len = progData.length() + varHeader.length();
		varEnt += LOBYTE( len );
		varEnt += HIBYTE( len );
		varEnt += 0x05;
		for( int y = 0; y < 8; y++ )
			varEnt += ( ProgName[y] == '/' ? '\0' : ProgName[y] );
		//varEnt += "TESTPROG";
		varEnt += '\0';
		varEnt += '\0';
		varEnt += LOBYTE( len );
		varEnt += HIBYTE( len );

		// now setup the file header
		string filHeader = "";
		filHeader += "**TI83F*";
		filHeader += 0x1A;
		filHeader += 0x0A;
		filHeader += '\0';
		for( int z = 0; z < 42; z++ )
			filHeader += '\0';
		len = varEnt.length() + varHeader.length() + progData.length();
		filHeader += LOBYTE( len );
		filHeader += HIBYTE( len );

		// now build the file itself
		string fileData = "";
		fileData += filHeader;
		fileData += varEnt;
		fileData += varHeader;
		fileData += progData;

		// tell the user what's happening
		DisplayInfo( "Calculating file checksum..." );

		// calculate the checksum
		string dataSect = "";
		dataSect += varEnt;
		dataSect += varHeader;
		dataSect += progData;
		fileData += progChecksum( (unsigned char*) dataSect.c_str(), dataSect.length() );

		// full filename
		string destFileName = destPath + destFile;

		// open the output file
		FILE* fp = fopen( destFileName.c_str(), "wb" );

		// check for a valid pointer
		if( ! ( fp ) )
		{
			// tell the user
			errstr = "Couldn't open output file!";
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}

		// tell the user what's happening
		DisplayInfo( "Writing to file..." );

		// write the data
		fwrite( fileData.c_str(), fileData.length(), 1, fp );

		// close the file
		fclose( fp );

		// close the input file
		ifs.close();

		// we're done
		cout << "Compilation complete!" << endl;

	} // if( Compiling )
	else // if( ! ( Compiling ) )
	{
		// tell the user
		cout << "Decompiling " << argv[1] << endl;

		// the output filename
		string outFile = RemoveExtension( fileTitle, ".8xp" ) + ".tib";

		// the destination path
		string outPath = RemoveFilename( argv[1] );

		// check if it's valid
		if( outFile == "..tib" )
		{
			// tell the user
			errstr = "Invalid filename: " + string( argv[1] );
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}
		
		// open the file
		FILE* fp = fopen( argv[1], "rb" );

		// check that it's valid
		if( fp == NULL )
		{
			// tell the user
			errstr = "Couldn't open " + string( argv[1] );
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}

		// structures for use later
		struct PrgmHeader head;
		struct VarEnt vent;

		// read in the program's header
		fread( &head, sizeof( PrgmHeader ), 1, fp );

		// read in the variable entry
		fread( &vent, sizeof( VarEnt ), 1, fp );

		// inform the user of what's happening
		DisplayInfo( "Reading tokens..." );

		// temp storage
		unsigned char tmp[2];

		// number of bytes read
		unsigned short numbyts = 0;

		// set the program data to nothing
		progData = "";

		// main read loop
		while( ( ! ( feof( fp ) ) ) && numbyts < vent.VarLength2 )
		{
			// read in a byte
			fread( tmp, sizeof( unsigned char ), 1, fp );

			// increment the number of bytes
			numbyts++;

			// check for assembly
			if( assembly )
			{
				// check for newline
				if( tmp[0] == 0x3F )
				{
					// append a colon and a newline
					progData += "\n";
				}
				// no newline
				else
				{
					progData += tmp[0];
				}

				// skip the replacements
				continue;
			}

			// check for a newline
			if( tmp[0] == 0x3F )
			{
				// append a colon and a newline
				progData += "\n";

				// skip the checks
				continue;
			}

			// index
			int i;

			// search the two-byte stuff first (special handling here)
			for( i = 0; i < numVars; i++ )
			{
				// check
				if( LOBYTE( CalcVars[i].data ) == tmp[0] )
				{
					// save the old data
					char tmp_sav[2];

					// copy
					memcpy( tmp_sav, tmp, 1 );

					// read in a byte
					fread( tmp, sizeof( unsigned char ), 1, fp );

					// increment the number of bytes
					numbyts++;

					// validate it
					if( HIBYTE( CalcVars[i].data ) == tmp[0] )
					{
						// check if we need to do assembly
						if( strcmp( CalcVars[i].text, "AsmPrgm" ) == 0 )
						{
							assembly = true;
						}

						// append the string
						progData += CalcVars[i].text;

						// break from the loop
						break;
					}

					// seek back 1
					fseek( fp, -1, SEEK_CUR );

					// less bytes read
					numbyts--;

					// restore the old data
					memcpy( tmp, tmp_sav, 1 );
				}
			}

			// check if a match was made
			if( i != numVars )
				continue;

			// try out the standard tokens
			for( i = 0; i < numTokens; i++ )
			{
				// check
				if( tmp[0] == StandardTokens[i].data )
				{
					// append the data
					progData += StandardTokens[i].text;

					// end the loop
					break;
				}
			}

			// check if a match was made
			if( i != numTokens )
				continue;

			// finally check for a replacement
			for( i = 0; i < numReps; i++ )
			{
				// check
				if( tmp[0] == Replacements[i].tok )
				{
					// append the character
					progData += Replacements[i].c;

					// break free
					break;
				}
			}

			// check if a match was made
			if( i != numReps )
				continue;

			// just append the character
			progData += (char) tmp[0];
		}

		// close the file
		fclose( fp );

		// make the filename
		string outFileName = outPath + outFile;

		// inform the user of what's happening
		DisplayInfo( "Writing output file..." );

		// open the output file
		ofstream ofs;
		ofs.open( outFileName.c_str() );

		// check that it's valid
		if( ofs.bad() || ofs.fail() )
		{
			// tell the user
			errstr = "Couldn't open output file!";
			DisplayError( const_cast< char* >( errstr.c_str() ) );

			// free the memory we took
			free( fileTitle );

			// return an error
			return -1;
		}
		
		// write the data
		ofs << progData;

		// close the output file
		ofs.close();

		// all done!
		cout << "Decompilation complete!" << endl;
	}

	// free the memory we took
	free( fileTitle );

	// return success
	return 0;
}
